"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CxxVector = exports.CxxVectorType = exports.Structure = exports.Float64 = exports.Float32 = exports.Int32 = exports.Int16 = exports.Int8 = exports.Uint32 = exports.Uint16 = exports.Uint8 = exports.Primitive = exports.TypedPointer = exports.CxxString = void 0;
const native_1 = require("./native");
Object.defineProperty(exports, "Primitive", { enumerable: true, get: function () { return native_1.Primitive; } });
const Uint8 = native_1.Primitive.Uint8;
exports.Uint8 = Uint8;
const Uint16 = native_1.Primitive.Uint16;
exports.Uint16 = Uint16;
const Uint32 = native_1.Primitive.Uint32;
exports.Uint32 = Uint32;
const Int8 = native_1.Primitive.Int8;
exports.Int8 = Int8;
const Int16 = native_1.Primitive.Int16;
exports.Int16 = Int16;
const Int32 = native_1.Primitive.Int32;
exports.Int32 = Int32;
const Float32 = native_1.Primitive.Float32;
exports.Float32 = Float32;
const Float64 = native_1.Primitive.Float64;
exports.Float64 = Float64;
exports.CxxString = {
    get(ptr) {
        return ptr.getCxxString();
    },
    set(ptr, value) {
        ptr.setCxxString(value);
    }
};
native_1.NativePointer.size = 8;
native_1.NativePointer.get = function (ptr) {
    return new native_1.NativePointer(ptr);
};
native_1.NativePointer.set = function (ptr, value) {
    value.setPointer(ptr);
};
native_1.SharedPointer.size = 16;
native_1.SharedPointer.get = function (ptr) {
    return new native_1.SharedPointer(ptr);
};
native_1.SharedPointer.set = function (ptr, value) {
    value.assignTo(ptr);
};
native_1.NetworkIdentifier.size = 152;
native_1.NetworkIdentifier.get = native_1.NetworkIdentifier.fromPointer;
native_1.NetworkIdentifier.set = function (ptr, value) {
    value.assignTo(ptr);
};
native_1.Actor.get = function (ptr) {
    return native_1.Actor.fromPointer(ptr.getPointer());
};
native_1.Actor.set = function (ptr, value) {
    ptr.setPointer(value);
};
class TypedPointer {
    constructor(type, address) {
        this.type = type;
        this.address = address;
    }
    get() {
        return this.type.get(this.address);
    }
    set(value) {
        this.type.set(this.address, value);
    }
    static get(type, ptr) {
        return new TypedPointer(type, ptr.getPointer());
    }
    static set(ptr, value) {
        ptr.setPointer(value.address);
    }
}
exports.TypedPointer = TypedPointer;
class Structure {
    static get(ptr) {
        const instance = new this();
        let off = 0;
        for (const [offset, key, type] of this.structure) {
            ptr.move(offset - off);
            off = offset;
            instance[key] = type.get(ptr);
        }
        ptr.move(-off);
        return instance;
    }
    static set(ptr, instance) {
        let off = 0;
        for (const [offset, key, type] of this.structure) {
            ptr.move(offset - off);
            off = off;
            type.set(ptr, instance[key]);
        }
        ptr.move(-off);
    }
}
exports.Structure = Structure;
class CxxVectorType {
    constructor(type) {
        this.type = type;
        this.name = `std::vector<${type.name}>`;
        if (this.type.size === undefined)
            throw Error("Cannot use unknown-sized type to CxxVector: " + type.name);
    }
    get(ptr) {
        return new CxxVector(ptr, this.type);
    }
    set(ptr, value) {
        console.assert(ptr.equals(value.address));
    }
}
exports.CxxVectorType = CxxVectorType;
class CxxVector {
    constructor(address, type) {
        this.address = address;
        this.type = type;
        if (!this.type.size)
            throw Error("CxxVector type needs size");
        this.ctor = this.type.ctor || (() => { });
        this.dtor = this.type.dtor || (() => { });
        if (this.type.ctor_move) {
            this.ctor_move = this.type.ctor_move;
        }
        else if (this.type.ctor_copy) {
            this.ctor_move = (to, from) => {
                this.type.ctor_copy(to, from);
                this.dtor(from);
            };
        }
        else {
            this.ctor_move = (to, from) => {
                this.ctor(to);
                this.type.set(to, this.type.get(from));
                this.dtor(from);
            };
        }
    }
    _resize(newsizeBytes, newcapBytes, oldptr, oldsizeBytes) {
        const compsize = this.type.size;
        const allocated = native_1.std$_Allocate$16(newsizeBytes);
        this.address.setPointer(allocated, 0);
        const oldptrmove = new native_1.NativePointer(oldptr);
        const oldsize = oldsizeBytes / compsize | 0;
        for (let i = 0; i < oldsize; i++) {
            this.ctor_move(allocated, oldptrmove);
            allocated.move(compsize);
            oldptrmove.move(compsize);
        }
        native_1.free(oldptr);
        this.address.setPointer(allocated, 8);
        allocated.move(newcapBytes - newsizeBytes);
        this.address.setPointer(allocated, 16);
    }
    set(idx, component) {
        const compsize = this.type.size;
        let begptr = this.address.getPointer(0);
        const oldsizeBytes = this.address.getPointer(8).subptr(begptr);
        if (idx * compsize < oldsizeBytes) {
            begptr.move(idx * compsize);
            this.type.set(begptr, component);
            return;
        }
        const cap = this.address.getPointer(16).subptr(begptr);
        if (idx >= cap) {
            const newsize = idx + 1;
            this._resize(newsize * compsize, Math.max(cap * 2, newsize), begptr, oldsizeBytes);
            begptr = this.address.getPointer(0);
        }
        begptr.move(oldsizeBytes);
        for (let i = oldsizeBytes / compsize | 0; i < idx; i++) {
            this.ctor(begptr);
            begptr.move(compsize);
        }
        this.ctor(begptr);
        this.type.set(begptr, component);
    }
    get(idx) {
        const beginptr = this.address.getPointer(0);
        const endptr = this.address.getPointer(8);
        const compsize = this.type.size;
        const size = endptr.subptr(beginptr) / compsize | 0;
        if (idx >= size)
            return null;
        beginptr.move(idx * compsize);
        return this.type.get(beginptr);
    }
    pop() {
        const beginptr = this.address.getPointer(8);
        const endptr = this.address.getPointer(16);
        if (endptr.equals(beginptr))
            return null;
        endptr.move(-this.type.size);
        const out = this.type.get(endptr);
        this.dtor(endptr);
        this.address.setPointer(endptr, 8);
        return out;
    }
    push(component) {
        let endptr = this.address.getPointer(8);
        const capptr = this.address.getPointer(16);
        if (capptr.equals(endptr)) {
            const begptr = this.address.getPointer(0);
            const oldsizeBytes = endptr.subptr(begptr);
            const capBytes = capptr.subptr(begptr);
            const newsizeBytes = oldsizeBytes + this.type.size;
            this._resize(newsizeBytes, Math.max(capBytes * 2, newsizeBytes), begptr, oldsizeBytes);
            endptr = this.address.getPointer(8);
        }
        this.ctor(endptr);
        this.type.set(endptr, component);
        endptr.add(this.type.size);
        this.address.setPointer(endptr, 8);
    }
    size() {
        const beginptr = this.address.getPointer(0);
        const endptr = this.address.getPointer(8);
        return endptr.subptr(beginptr) / this.type.size | 0;
    }
}
exports.CxxVector = CxxVector;
//# sourceMappingURL=nativetype.js.map